import copy
import itertools
from os.path import normpath
from pathlib import Path
from typing import Any, Callable, List

import pystache
from fspathtree import fspathtree

from . import expressions, graphs, parsing, units

__local_Quantity = units.Q_


def expand_partial_configs(configs: List[fspathtree], include_base=False):
    """
    Give a list of configuration trees, treat the frist tree
    as a "base" configuration and all other trees as partial configurations
    that should be merged with the base config to create instances.

    @param include_base : if true, the base config will be included in the returned set.
    """
    assert len(configs) > 0

    if len(configs) == 1:
        return configs

    full_configs = []
    if include_base:
        full_configs.append(copy.deepcopy(configs[0]))
    for c in configs[1:]:
        full_configs.append(copy.deepcopy(configs[0]))
        full_configs[-1].update(c)

    return full_configs


def try_construct_quantity(
    obj: Any,
    is_quantity: Callable[[Any], bool] = None,
    quantity_class: Any = __local_Quantity,
):
    """
    Try to construct a pint.UnitRegistry().Quantity
    object from a given object, if the given looks like it would be
    a quantity.

    By default, an object is considered a quantity if it:
        - is a string,
        - begins with a numerical value,
        - does not throw an exception when passed to pint.UnitRegistry().Quantity(...) constructor

    @param is_quantity: a function returning true if the object should be considered a quantity. This allows the caller to provide
                        their logic to identify quantities.
    @param quantity_class : the class that should be used to construct quantities. pint cannot handle quantities from different
                            registries, so this allows the caller to provide the class they are uing.
    """

    if is_quantity is None:

        def f(obj: Any) -> bool:
            if type(obj) is not str:
                return False
            try:
                float(obj.strip().split(" ")[0])
            except Exception:
                return False
            try:
                quantity_class(obj.strip())
            except Exception:
                return False

            return True

        is_quantity = f

    if is_quantity(obj):
        return quantity_class(obj)
    else:
        return obj


def expand_variables(text: Any, template: str = "ctx['{name}']"):
    """
    Expand shell-style variables into python variables

    @param template: optional template string used for variable expantion. The variables
                     name will be inserted into the template using {VARNAME}. Default value
                     is "ctx['{name}']", which means we will have:

                    $x -> c['x']
                    ${x} -> c['x']
                    $/grid/x -> c['/grid/x']
                    ${/grid/x} -> c['/grid/x']

    """
    i = 0
    new_text = ""
    for tokens, start, end in parsing.variable.scan_string(text):
        new_text += text[i:start]
        i = end
        new_text += template.format(name=tokens["variable name"])
    new_text += text[i:]
    return new_text


def contains_expression(text: Any):
    """
    Returns true if a string contains expression.
    Supports passing not strings as arguments, in which
    case False is returned.
    """

    if type(text) is not str:
        return False

    results = parsing.expression.search_string(text)
    return len(results) > 0


def contains_variable(text: Any):
    """
    Returns true if a string contains a variable refernce.
    Supports passing not strings as arguments, in which
    case False is returned.
    """
    if type(text) is not str:
        return False

    results = parsing.variable.search_string(text)
    return len(results) > 0


def load_includes(config: fspathtree, loader):
    """
    Find all '@include' branches in the tree and load the branch from
    a file.

    @param config: the configuration tree to expand. If the tree does not include any '@include' leafs, it will not be changes.
    @param loader: a function that accepts a pathlib.Path argument and loads the contents of the file into an fspathtree that is returned.
    """
    for leaf in list(config.get_all_paths(predicate=lambda p: p.name == "@include")):
        if type(config[leaf]) == str:
            loaded_config = loader(Path(config[leaf]))
            loaded_config = load_includes(loaded_config, loader)
        elif type(config[leaf]) == fspathtree and type(config[leaf].tree) == list:
            # load first include
            loaded_config = loader(Path(config[leaf][0]))
            # load others using update method
            for i in range(1, len(config[leaf])):
                loaded_config.update(loader(Path(config[leaf][i])))
            loaded_config = load_includes(loaded_config, loader)
        else:
            raise RuntimeError(
                "'@include' node value must be a string or a list of strings."
            )

        if len(leaf.parts) == 2:  # an @include at the config root
            # we want any keys in the main config to override the config
            # loaded from file
            loaded_config.update(config)
            config = loaded_config
        else:
            config[leaf.parent] = loaded_config.tree

    return config


class ConfigRenderer:
    Quantity = units.Q_

    def __init__(self, expression_evaluator=None):
        if expression_evaluator is None:
            expression_evaluator = expressions.ExecExpressionEvaluator()

        self.expression_evaluator = expression_evaluator

    def expand_and_render(self, config):
        """Expand batch configurations and render each instance."""
        configs = self.expand_batch_nodes(config)
        for i in range(len(configs)):
            configs[i] = self.render(configs[i])

        return configs

    def expand_batch_nodes(self, config: fspathtree):
        """Expand @batch nodes in a configuration tree into multiple configuration trees."""
        configs = []

        batch_leaves = list(
            map(
                lambda p: str(p.parent),
                config.get_all_paths(predicate=lambda p: p.name == "@batch"),
            )
        )

        for vals in itertools.product(
            *[config[leaf + "/@batch"] for leaf in batch_leaves]
        ):
            instance = copy.deepcopy(config)
            for i, leaf in enumerate(batch_leaves):
                instance[leaf] = vals[i]
            configs.append(instance)

        return configs

    def render(self, config: fspathtree, make_copy=True):
        """Render a configuration tree, constructing all quantities, expanding all variables, and evaluating all expressions."""
        # create a graph of the dependencies
        G = graphs.DependencyGraph()
        # add all leaf node paths as nodes of the graph
        G.add_nodes_from(config.get_all_leaf_node_paths())
        # draw edges between nodes representing dependencies.
        # edges point from a node to its dependency.
        # if a node has incomming edges, other nodes depend on it
        # if a node has outgoing edges, it depends on others.
        for node in list(G.nodes):
            variables = [
                fspathtree.PathType(r["variable name"])
                for r in parsing.variable.search_string(config[node])
            ]
            for v in variables:
                dep = fspathtree.PathType(
                    normpath(v if v.is_absolute() else node.parent / v)
                )
                G.add_edge(node, dep)
        # detect circular dependencies
        cycles = sorted(graphs.nx.simple_cycles(G))
        if len(cycles) > 0:
            msg = "Circular dependencies detected."
            for cycle in cycles:
                msg += "(" + " -> ".join(map(str, cycle)) + ")"
            raise RuntimeError(msg)

        # make a copy to work with

        if make_copy:
            rendered_config = copy.deepcopy(config)
        else:
            rendered_config = config

        rendered_config = self._construct_all_quantities(rendered_config)
        rendered_config = self._expand_all_variables(rendered_config)
        rendered_config = self._evaluate_all_expressions(rendered_config, G)

        return rendered_config

    def _evaluate_all_expressions(
        self, config: fspathtree, graph: graphs.DependencyGraph
    ):
        """Evaluate the expressions in a configuration tree, using a graph of the tree dependencies to determine the render order."""
        # We need to determine the order to evaluate the nodes of the config tree.
        # We have a graph that describes the dependencies. Each node is the graph is a leaf node
        # in the tree, and the edges represent dependencies between nodes.
        # Consider an example,
        #
        #          A        X
        #        /   \    /   \
        #       B    C   Y     Z
        #           / \ /
        #          D   E
        #
        # Assume all edges are directed DOWN.
        # B, D, E and Z then have no dependencies. All of their edges are "in" edges,
        # they have no "out" edges.
        #
        # With networkx we can easily get a list of all nodes with no dependencies, first
        # get the number of in and out edges for each node.
        out_degrees = dict(graph.out_degree)
        in_degrees = dict(graph.in_degree)
        # any nodes that have zero out edges (out_degrees = 0) have no dependencies, and we can go ahead and render these.
        config = self._evaluate_expressions(
            config,
            paths=map(
                lambda item: item[0],
                filter(lambda item: item[1] == 0, out_degrees.items()),
            ),
        )
        # Now we need to render nodes that have dependencies, but we need to determine the order that this can be
        # done first. In the above example, we can't evaluate A until B and C are evaluated. We can't evaluate C until D and E
        # are evaluated.
        #
        # With networkx, we can easily get a list of all ancestors of a node. For node E, this would be [C,Y,A,X]
        # Note: with the direction of our edges, an ancestor _depends_ on the node. So C is an ancestor of E, even
        # though it seems more natural to consider E the ancestor of C. If we wanted E to be the ancestor of C,
        # we would need to direct our edges to point from dependencies to dependents.

        # Find all root dependencies. those nodes that don't depend on any others, but are depended on by others
        # thse are all nodes with out_degrees == 0 and in_degrees > 0
        root_dependencies = list(
            filter(
                lambda k: in_degrees[k] > 0,
                filter(lambda k: out_degrees[k] == 0, out_degrees.keys()),
            )
        )

        # For each root node, get a list of all ancestors, and then find every path that connects the ancestor to the root.
        # in this example above, node E would have 4 ancestors, [C,Y,A,X]. And the list of paths connecting E to its ancestors
        # would be E -> C, E -> Y, E -> C -> A, and E -> Y -> X.
        # However, we can't simply take one of these paths and start evaluating nodes. If we tried to evaluate E -> C -> A for example,
        # we would get an error if B had not already been determined.
        #
        # What we really need to be able to do is find A and X, then work backward from there...
        #
        # Find all leaf node dependencies. These are nodes that only have dependencies, no dependents (out_degree > 0, in_nodes = 0)
        leaf_dependencies = list(
            filter(
                lambda k: in_degrees[k] == 0,
                filter(lambda k: out_degrees[k] > 0, out_degrees.keys()),
            )
        )
        # now get a set of paths that connect each leaf node to its children
        for leaf in leaf_dependencies:
            all_paths = []
            for d in graphs.nx.descendants(graph, leaf):
                all_paths += graphs.nx.all_simple_paths(graph, leaf, d)

            # node batches are sets of nodes that can all be rendered
            # at the same time.
            node_batches = [
                set(filter(lambda n: n is not None, nodes))
                for nodes in itertools.zip_longest(*all_paths)
            ]
            node_batches.reverse()
            for batch in node_batches[1:]:
                config = self._evaluate_expressions(config, paths=batch)

        return config

        for root in root_dependencies:
            all_chains = []
            for a in graphs.nx.ancestors(graph, root):
                all_chains += graphs.nx.all_simple_paths(graph, a, root)

        # get all paths in the graph that end on roots
        all_chains = []
        for root in root_dependencies:
            for a in graphs.nx.ancestors(graph, root):
                all_chains += graphs.nx.all_simple_paths(graph, a, root)
        # prune paths that are included in ohters,
        # i.e. we only want to keep paths from each root to it's
        # oldest ancestors.
        longest_chains = []
        for p1 in all_chains:
            save = True
            for p2 in all_chains:
                if p1 == p2:
                    continue
                if p2[-len(p1) :] == p1:
                    save = False
                    break
            if save:
                longest_chains.append(p1)

        # determine the order to evaluate dependencies.

        for chain in longest_chains:
            chain.reverse()  # paths start with oldest ancestor and end with root node.
            config = self._evaluate_expressions(config, paths=chain[1:])

        return config

    def _evaluate_expressions(self, config: fspathtree, paths: List[Any]):
        """Evaluate expression in the config at the paths listed. Uses as a utility function on _evaluate_all_expressions(...)"""
        for path in paths:
            if not contains_expression(config[path]):
                continue

            self.expression_evaluator.globals["ctx"] = config[path.parent]
            expressions = parsing.expression.search_string(config[path])
            if (
                len(expressions) == 1
                and "$" + expressions[0]["expression body"] == config[path]
            ):
                # the value of the element is a single expression with no surrounding text
                # we want to replace the expression with the evaluation
                e = expressions[0]
                value = self.expression_evaluator.eval(e["expression body"][1:-1])
                config[path] = value
            else:
                # we have more than one expression or the expression is surrounded by text
                # we want to evaluate each expression and replace it with a str of its value
                old_text = config[path]
                new_text = ""
                i = 0
                for tokens, start, end in parsing.expression.scan_string(old_text):
                    new_text += old_text[i:start]
                    i = end
                    new_text += str(
                        self.expression_evaluator.eval(tokens["expression body"][1:-1])
                    )
                new_text += old_text[i:]
                config[path] = new_text
        return config

    def _construct_all_quantities(self, config: fspathtree):
        """Replace strings in the tree representing quantities with pint.Quantity objects."""
        for path in config.get_all_leaf_node_paths():
            config[path] = try_construct_quantity(
                config[path], quantity_class=self.Quantity
            )
        return config

    def _expand_all_variables(
        self, config: fspathtree, template: str = "ctx['{name}']"
    ):
        """
        Expand all shell-style variables into python variables in the entire tree.
        """
        for path in config.get_all_leaf_node_paths():
            if type(config[path]) is str:
                config[path] = expand_variables(config[path])

        return config


def render_mustache_template(template_text: str, ctx: fspathtree):
    """
    Render a Mustache template text string using the values in a context. The context can
    be a configuration tree that has been previously loaded and rendered with powerconf.
    """
    # check for keys in the template that are not in the config dict
    # and throw an exception if any are missing
    parse_tree = pystache.parse(template_text)
    missing_keys = list(
        filter(
            lambda item: item not in ctx,
            set(
                map(
                    lambda item: item.key,
                    filter(lambda item: type(item) is not str, parse_tree._parse_tree),
                )
            ),
        )
    )
    if len(missing_keys) > 0:
        msg = (
            "Required configuration parameters were missing. These parameters are present in the template, but were not given in the context: "
            + ", ".join(missing_keys)
        )
        raise RuntimeError(msg)

    # we have to pass a flat dict to pystache, so we generate
    # it inline here. also, Mustache uses a leading slash on a key to
    # identify the end of a section, so we can use absolute path, just
    # relative paths with respect to the root.
    rendered_text = pystache.render(
        template_text, {str(p)[1:]: ctx[p] for p in ctx.get_all_leaf_node_paths()}
    )

    return rendered_text


def render_mustache_template_file(
    template_file: Path, ctx: fspathtree, output_file: Path
):
    template_text = template_file.read_text()
    rendered_text = render_mustache_template(template_text, ctx)
    output_file.write_text(rendered_text)



import pint
import pytest
from fspathtree import fspathtree

from powerconf import rendering

ureg = pint.UnitRegistry()
Q_ = ureg.Quantity


def test_quantity_construction():
    x = rendering.try_construct_quantity(" 4.5 cm")
    assert x.magnitude == pytest.approx(4.5)

    x = rendering.try_construct_quantity(4)
    assert x == 4

    x = rendering.try_construct_quantity(4.5)
    assert x == pytest.approx(4.5)

    x = rendering.try_construct_quantity("quant")
    assert x == "quant"

    x = rendering.try_construct_quantity("quad")
    assert x == "quad"
    x = rendering.try_construct_quantity("quad", is_quantity=lambda s: True)
    assert x.magnitude == pytest.approx(1)

    with pytest.raises(Exception):
        x = rendering.try_construct_quantity("lskjdf", is_quantity=lambda s: True)

    x = rendering.try_construct_quantity(" 4.5 cm")
    assert type(x) is not Q_
    x = rendering.try_construct_quantity(" 4.5 cm", quantity_class=Q_)
    assert type(x) is Q_


def test_mustache_template_rendering():
    ctx = fspathtree({"grid": {"x": {"max": "1 cm"}}})

    template_text = "grid.x.max = {{grid/x/max}}"
    rendered_text = rendering.render_mustache_template(template_text, ctx)
    assert rendered_text == "grid.x.max = 1 cm"

    template_text = "grid.x.max = {{grid/y/max}}"
    with pytest.raises(RuntimeError) as e:
        rendered_text = rendering.render_mustache_template(template_text, ctx)
    assert "Required configuration" in str(e)
    assert "grid/y/max" in str(e)



from os.path import normpath

# import matplotlib.pyplot as plt
import networkx as nx
from fspathtree import fspathtree

from powerconf import parsing


def test_creating_graph():
    """This is just a function for experimenting with the networkx library."""
    config = fspathtree(
        {
            "grid": {
                "res": "1 um",
                "x": {
                    "max": "1 cm",
                    "min": "0 cm",
                    "res": "${../res}",
                    "N": "($max - $min)/$res + 1",
                },
                "y": {
                    "max": "1 cm",
                    "min": "0 cm",
                    "res": "${../res}",
                    "N": "($max - $min)/$res + 1",
                },
            }
        }
    )

    G = nx.DiGraph()
    G.add_nodes_from(config.get_all_leaf_node_paths())
    for node in list(G.nodes):
        variables = [
            fspathtree.PathType(r["variable name"])
            for r in parsing.variable.search_string(config[node])
        ]
        for v in variables:
            dep = fspathtree.PathType(
                normpath(v if v.is_absolute() else node.parent / v)
            )
            G.add_edge(node, dep)

    # out_degrees are the number of other nodes a given node depends on.
    # in_degrees are the number of other nodes that depend on given node.
    # print()
    # print("out")
    # for e in G.out_degree:
    #     print(e)
    # print("in")
    # for e in G.in_degree:
    #     print(e)

    out_degrees = dict(G.out_degree)
    in_degrees = dict(G.in_degree)
    # at the bottom of a dependency tree is a node that does not depend on anything (out_degree == 0)
    # and has one or more nodes that depends on it (in_node > 0)
    root_dependencies = list(
        filter(
            lambda k: in_degrees[k] > 0,
            filter(lambda k: out_degrees[k] == 0, out_degrees.keys()),
        )
    )

    # print(">>>>", root_dependencies[0])
    # print(":::::", sorted(nx.ancestors(G, root_dependencies[0])))

    all_dep_chains = []
    for root in root_dependencies:
        for node in nx.ancestors(G, root):
            all_dep_chains += nx.all_simple_paths(G, node, root)
    uniq_dep_chains = []
    for chain1 in all_dep_chains:
        contained = False
        for chain2 in all_dep_chains:
            if len(set(chain1).intersection(set(chain2))) == len(chain1):
                contained = True
        if not contained:
            uniq_dep_chains.append(chain1)
    # print(">>>", all_dep_chains)
    # print("<<<", uniq_dep_chains)

    # evaluate any expressions in parameters that don't depend on any others
    for node in list(map(lambda e: e[0], filter(lambda e: e[1] == 0, G.out_degree))):
        # eval....
        pass

    for node in root_dependencies:
        for nnode in nx.algorithms.traversal.depth_first_search.dfs_postorder_nodes(
            G, node
        ):
            pass

    # nx.draw(G, with_labels=True)
    # plt.show()


def test_find_all_paths_ending_on_node():
    G = nx.DiGraph()

    G.add_node("a")
    G.add_node("b")
    G.add_node("c")
    G.add_node("d")

    G.add_edge("a", "b")
    G.add_edge("a", "c")
    G.add_edge("b", "d")
    G.add_edge("c", "d")

    all_paths = []
    for a in nx.ancestors(G, "d"):
        all_paths += nx.all_simple_paths(G, a, "d")

    # prune paths that are parts of longer paths
    longest_paths = []
    print()
    for p1 in all_paths:
        save = True
        for p2 in all_paths:
            if p1 == p2:
                continue
            if p2[-len(p1) :] == p1:
                save = False
                break
        if save:
            longest_paths.append(p1)

    assert len(longest_paths) == 2
    assert ["a", "b", "d"] in longest_paths
    assert ["a", "c", "d"] in longest_paths



import pyparsing as pp

unbraced_variable = pp.Combine(
    pp.Literal("$") + pp.Word(pp.alphanums + "/_")("variable name")
)
braced_variable = pp.Combine(
    pp.Literal("$")
    + pp.QuotedString(quote_char="{", end_quote_char="}")("variable name")
)
variable = unbraced_variable | braced_variable
expression = pp.Combine(
    pp.Literal("$") + pp.original_text_for(pp.nested_expr())("expression body")
)



import pathlib

import pint
import pytest
import yaml
from fspathtree import fspathtree

from powerconf import expressions, loaders, rendering, units, utils

from . import unit_test_utils

ureg = pint.UnitRegistry()
Q_ = ureg.Quantity


def test_batch_expansion_single_node():
    config = fspathtree({"time": {"max": {"@batch": [1, 2, 3]}}})
    config_renderer = rendering.ConfigRenderer(expressions.ExecExpressionEvaluator())

    configs = config_renderer.expand_batch_nodes(config)
    assert len(configs) == 3
    assert configs[0]["/time/max"] == 1
    assert configs[1]["/time/max"] == 2
    assert configs[2]["/time/max"] == 3


def test_batch_expansion_two_nodes():
    config = fspathtree(
        {
            "time": {"max": {"@batch": [1, 2, 3]}},
            "grid": {"x": {"max": {"@batch": [4, 5]}}},
        }
    )
    config_renderer = rendering.ConfigRenderer(expressions.ExecExpressionEvaluator())

    configs = config_renderer.expand_batch_nodes(config)
    assert len(configs) == 6
    assert configs[0]["/time/max"] == 1
    assert configs[1]["/time/max"] == 1
    assert configs[2]["/time/max"] == 2
    assert configs[3]["/time/max"] == 2
    assert configs[4]["/time/max"] == 3
    assert configs[5]["/time/max"] == 3
    assert configs[0]["/grid/x/max"] == 4
    assert configs[1]["/grid/x/max"] == 5
    assert configs[2]["/grid/x/max"] == 4
    assert configs[3]["/grid/x/max"] == 5
    assert configs[4]["/grid/x/max"] == 4
    assert configs[5]["/grid/x/max"] == 5


def test_expand_to_list():
    config = fspathtree({"N": 10, "xmin": -1, "xmax": 1, "points": "$([1,2,3])"})
    config_renderer = rendering.ConfigRenderer()
    config = config_renderer.render(config)

    assert config["N"] == 10
    assert type(config["points"].tree) is list
    assert config["points"][0] == 1
    assert config["points"][1] == 2
    assert config["points"][2] == 3

    # list comprehension
    config = fspathtree(
        {
            "N": 10,
            "xmin": -1,
            "xmax": 1,
            "dx": "$((${xmax}-${xmin})/(${N}-1) )",
            "points": "$([ i*${dx} + ${xmin} for i in range(${N})])",
        }
    )
    config_renderer = rendering.ConfigRenderer()
    config = config_renderer.render(config)

    assert config["N"] == 10
    assert type(config["points"].tree) is list
    assert len(config["points"].tree) == 10
    assert config["points"][0] == -1
    assert config["points"][1] == pytest.approx(-1 + 2 / 9)
    assert config["points"][2] == pytest.approx(-1 + 4 / 9)
    assert config["points"][8] == pytest.approx(1 - 2 / 9)
    assert config["points"][9] == 1

    # list comprehension with quantities
    config = fspathtree(
        {
            "N": 10,
            "xmin": "-1 cm",
            "xmax": "1 cm",
            "dx": "$((${xmax}-${xmin})/(${N}-1) )",
            "points": "$([ i*${dx} + ${xmin} for i in range(${N})])",
        }
    )
    config_renderer = rendering.ConfigRenderer()
    config = config_renderer.render(config)

    assert config["N"] == 10
    assert type(config["points"].tree) is list
    assert len(config["points"].tree) == 10
    assert config["points"][0].to("m").magnitude == pytest.approx(-0.01)
    assert config["points"][1].to("m").magnitude == pytest.approx(-0.01 + 0.02 / 9)
    assert config["points"][2].to("m").magnitude == pytest.approx(-0.01 + 0.04 / 9)
    assert config["points"][8].to("m").magnitude == pytest.approx(0.01 - 0.02 / 9)
    assert config["points"][9].to("m").magnitude == pytest.approx(0.01)

    # list comprehension with quantities
    config = fspathtree(
        {
            "N": 10,
            "t0": "200 us",
            "tau": "50 us",
            "pulses": "$([ {'arrival_time': i*${t0}, 'duration' : ${tau}} for i in range(${N})])",
        }
    )
    config_renderer = rendering.ConfigRenderer()
    config = config_renderer.render(config)

    assert config["N"] == 10
    assert type(config["pulses"].tree) is list
    assert len(config["pulses"].tree) == 10
    assert config["pulses/0/arrival_time"] == 0
    assert config["pulses/1/arrival_time"].magnitude == 200
    assert config["pulses/2/arrival_time"].magnitude == 400


def test_expression_detector():
    assert not rendering.contains_expression(1)
    assert not rendering.contains_expression("x")
    assert not rendering.contains_expression("${x}")
    assert rendering.contains_expression("$(1 + 1)")
    assert rendering.contains_expression("_$(1 + 1)-var")
    assert rendering.contains_expression("_$(1 + 1)-to-$(2+2)")


def test_evaluate_expression():
    config = fspathtree(
        {
            "time": {"max": "$(1 + 2)"},
            "prefix": "_$(1 + 2)",
            "prefix2": "_$(1 + 2)-$(2+3)_",
            "sin_of_1": "$(math.sin(1))",
            "grid": {
                "x": {
                    "min": Q_(0, "cm"),
                    "length": Q_(1, "cm"),
                    "n": "$(2*ctx['../n'])",
                    "max": "$(ctx['min']+ctx['length'])",
                },
                "n": 10,
            },
        }
    )
    evaluator = expressions.ExecExpressionEvaluator()
    import math

    evaluator.add_global("math", math)
    config_renderer = rendering.ConfigRenderer(evaluator)

    config = config_renderer.render(config)

    assert config["/time/max"] == 3
    assert config["/prefix"] == "_3"
    assert config["/prefix2"] == "_3-5_"
    assert config["/sin_of_1"] == pytest.approx(math.sin(1))
    assert config["/grid/x/min"].magnitude == pytest.approx(0)
    assert config["/grid/x/max"].magnitude == pytest.approx(1)


def test_expand_variables():
    assert rendering.expand_variables("${x}", "ctx['{name}']") == "ctx['x']"
    assert (
        rendering.expand_variables("${x} + $y", "ctx['{name}']")
        == "ctx['x'] + ctx['y']"
    )


def test_construct_quantities():
    config = fspathtree(
        {"time": {"max": "2 s"}, "laser": {"power": "1 W/cm^2"}, "tag": "CW", "N": 10}
    )
    config_renderer = rendering.ConfigRenderer()
    config = config_renderer._construct_all_quantities(config)

    assert config["/time/max"].magnitude == pytest.approx(2)
    assert config["/laser/power"].to("mW/cm^2").magnitude == pytest.approx(1000)
    assert config["/tag"] == "CW"
    assert config["/N"] == 10


def test_only_construct_quantities_from_strings_that_look_like_a_quantity():
    """
    Pint is willing to treat way more text as a quantity than we want to allow.

    For example, '$(${cm})' will be interpretted as a quantity. So will 'quant'.

    So we only want to try to interpret something as a quantity if it starts wit a numerical value.
    """

    assert rendering.try_construct_quantity("4 m").magnitude == 4

    assert units.Q_("cm") is not None
    assert units.Q_("cm").magnitude == 1
    assert units.Q_("$(${cm})") is not None
    assert units.Q_("$(${cm})").magnitude == 1
    assert rendering.try_construct_quantity("cm") == "cm"
    assert rendering.try_construct_quantity("$(${cm})") == "$(${cm})"


def test_yaml_config_example_1():
    config_text = """
grid:
    res: 1 um
    x:
      min: 0 cm
      max: 1.5 cm
      N: $( ($max - $min) / ${../res} + 1 )
    y:
      min: 0 cm
      max: 0.5 cm
      N: $( ($max - $min) / ${../res} + 1 )
"""

    config = fspathtree(yaml.safe_load(config_text))

    config_renderer = rendering.ConfigRenderer()

    config = config_renderer.render(config)
    assert config["grid/x/max"].magnitude == pytest.approx(1.5)
    assert config["grid/x/N"] == 15001


def test_yaml_config_example_2():
    """Adding unit support to legacy configs..."""

    config_text = """
x_q: 13 mm
x: $($x_q.to("cm").magnitude)
"""

    config = fspathtree(yaml.safe_load(config_text))

    evaluator = expressions.ExecExpressionEvaluator()
    config_renderer = rendering.ConfigRenderer(evaluator)
    config = config_renderer.render(config)

    assert config["x_q"].magnitude == pytest.approx(13)
    assert config["x"] == pytest.approx(1.3)


def test_yaml_config_example_3():
    """Adding unit support to legacy configs..."""

    config_text = """
x_q: 13 mm
x: $($x_q.to("cm").magnitude)
"""

    config = fspathtree(yaml.safe_load(config_text))

    evaluator = expressions.ExecExpressionEvaluator()
    config_renderer = rendering.ConfigRenderer(evaluator)

    config = config_renderer.render(config)

    assert config["x_q"].magnitude == pytest.approx(13)
    assert config["x"] == pytest.approx(1.3)


def test_inplace_vs_copy_render():
    config_text = """
x_q: 13 mm
x: $($x_q.to("cm").magnitude)
"""

    config = fspathtree(yaml.safe_load(config_text))

    config_renderer = rendering.ConfigRenderer()

    rconfig = config_renderer.render(config)

    assert rconfig["x_q"].magnitude == pytest.approx(13)
    assert rconfig["x"] == pytest.approx(1.3)
    assert config["x_q"] == "13 mm"
    assert config["x"] == '$($x_q.to("cm").magnitude)'

    rconfig = config_renderer.render(config, make_copy=False)

    assert rconfig["x_q"].magnitude == pytest.approx(13)
    assert rconfig["x"] == pytest.approx(1.3)
    assert config["x_q"].magnitude == pytest.approx(13)
    assert config["x"] == pytest.approx(1.3)


def test_circular_dependency_detection():
    """Adding unit support to legacy configs..."""

    config_text = """
a : $($g)
b : $($a)
c : $($b)
d : $($c)
e : $($d)
f : $($e)
g : $($f)
"""

    config = fspathtree(yaml.safe_load(config_text))
    evaluator = expressions.ExecExpressionEvaluator()
    config_renderer = rendering.ConfigRenderer(evaluator)

    with pytest.raises(RuntimeError) as e:
        config = config_renderer.render(config)
    assert "Circular dependencies detected" in str(e)


def test_long_dependency_chaings():
    """Adding unit support to legacy configs..."""

    config_text = """
a : 'here'
b : $($a)
c : $($b)
d : $($c)
e : $($d)
f : $($e)
g : $($f)
"""

    config = fspathtree(yaml.safe_load(config_text))
    evaluator = expressions.ExecExpressionEvaluator()
    config_renderer = rendering.ConfigRenderer(evaluator)

    config = config_renderer.render(config)

    assert config["/g"] == "here"


def test_batch_expansion_and_rendering():
    config_text = """
a : 
    '@batch':
        - 1
        - 2
        - 3
b : $($a)
c : $($b)
d : $($c)
e : $($d)
f : $($e)
g : $($f)
"""

    config = fspathtree(yaml.safe_load(config_text))
    config_renderer = rendering.ConfigRenderer()
    configs = config_renderer.expand_and_render(config)

    assert len(configs) == 3
    assert configs[0]["/g"] == 1


def test_rendering_partial_configs():
    text = """
one: 1
two: 2
---
three : 3
    """

    configs = loaders.yaml_all_docs(text)
    configs = rendering.expand_partial_configs(configs)

    assert len(configs) == 1

    assert "one" in configs[0]
    assert "two" in configs[0]
    assert "three" in configs[0]

    configs = loaders.yaml_all_docs(text)
    configs = rendering.expand_partial_configs(configs, include_base=True)

    assert len(configs) == 2

    assert "one" in configs[0]
    assert "two" in configs[0]
    assert "three" not in configs[0]
    assert "one" in configs[1]
    assert "two" in configs[1]
    assert "three" in configs[1]


def test_rendering_partial_configs_with_nested_trees():
    text = """
sim:
    grid:
        res : 1 um
        x:
            min: 0 cm
            max: 1 cm
        y:
            min: 0 cm
            max: 1 cm
---
sim:
    grid:
        x:
            N: $( ($max-$min)/($/sim/grid/res) + 1)
        y:
            N: $( ($max-$min)/($/sim/grid/res) + 1)
    """

    configs = loaders.yaml_all_docs(text)
    configs = rendering.expand_partial_configs(configs)

    assert len(configs) == 1

    assert "/sim/grid/x/N" in configs[0]
    assert "/sim/grid/x/min" in configs[0]
    assert "/sim/grid/x/max" in configs[0]
    assert "/sim/grid/y/N" in configs[0]
    assert "/sim/grid/y/min" in configs[0]
    assert "/sim/grid/y/max" in configs[0]


def test_rendering_complex_dependencies():
    config = fspathtree(
        {
            "laser": {
                "irradiance": "1 W/cm^2",
                "one_over_e2_diameter": "1 cm",
                "one_over_e_diameter": "$(${one_over_e2_diameter}/math.sqrt(2))",
                "one_over_e_area": "$(math.pi*${one_over_e_diameter}**2/4)",
                "power": "$(${irradiance} * ${one_over_e_area})",
            }
        }
    )
    config_renderer = rendering.ConfigRenderer()
    config = config_renderer.render(config)

    assert config["/laser/one_over_e_diameter"].magnitude == pytest.approx(1 / 2**0.5)
    assert config["/laser/power"].magnitude == pytest.approx(1 * 3.14159 / 8)


def test_include_branches_yaml(tmp_path):
    with unit_test_utils.working_directory(tmp_path):
        text1 = """
    simulation:
        grid:
            x:
                '@include' : grid.yml
            y:
                '@include' : grid.yml
    """

        text2 = """
    min: 0 cm
    max: 1 cm
    N: 101
        """

        pathlib.Path("grid.yml").write_text(text2)

        config = loaders.yaml(text1)
        config = rendering.load_includes(config, loaders.yaml)

        assert "simulation/grid/x/min" in config
        assert "simulation/grid/x/max" in config
        assert "simulation/grid/x/N" in config
        assert "simulation/grid/y/min" in config
        assert "simulation/grid/y/max" in config
        assert "simulation/grid/y/N" in config

        assert config["simulation/grid/x/min"] == "0 cm"
        assert config["simulation/grid/x/max"] == "1 cm"
        assert config["simulation/grid/x/N"] == 101
        assert config["simulation/grid/y/min"] == "0 cm"
        assert config["simulation/grid/y/max"] == "1 cm"
        assert config["simulation/grid/y/N"] == 101


def test_multi_level_include_branches_yaml(tmp_path):
    with unit_test_utils.working_directory(tmp_path):
        text1 = """
    simulation:
        grid:
            '@include' : grid.yml
    """

        text2 = """
    x: 
        '@include' : x-grid.yml
    y: 
        '@include' : x-grid.yml
    """
        text3 = """
    min: 0 cm
    max: 1 cm
    N: 101
        """
        text4 = """
    min: 1 cm
    max: 2 cm
    N: 201
        """

        pathlib.Path("grid.yml").write_text(text2)
        pathlib.Path("x-grid.yml").write_text(text3)
        pathlib.Path("y-grid.yml").write_text(text4)

        config = loaders.yaml(text1)
        config = rendering.load_includes(config, loaders.yaml)

        assert "simulation/grid/x/min" in config
        assert "simulation/grid/x/max" in config
        assert "simulation/grid/x/N" in config
        assert "simulation/grid/y/min" in config
        assert "simulation/grid/y/max" in config
        assert "simulation/grid/y/N" in config

        assert config["simulation/grid/x/min"] == "0 cm"
        assert config["simulation/grid/x/max"] == "1 cm"
        assert config["simulation/grid/x/N"] == 101
        assert config["simulation/grid/y/min"] == "0 cm"
        assert config["simulation/grid/y/max"] == "1 cm"
        assert config["simulation/grid/y/N"] == 101


def test_yaml_config_var_name_errors():
    config_text = """
layers:
    - name: layer 1
      thickness: 10 um
grid:
    x:
      max: $( ${../layers/0/thickness}) # element name does not walk up the tree far enough
"""

    config = fspathtree(yaml.safe_load(config_text))

    config_renderer = rendering.ConfigRenderer()

    with pytest.raises(KeyError) as e:
        config = config_renderer.render(config)

    assert (
        e.value.args[0]
        == "Error while parsing path '/grid/layers/0/thickness'. Could not find path element 'layers'."
    )

    config_text = """
layers:
    - name: layer 1
      thickness: 10 um
grid:
    x:
      max: $( ${/layer/0/thickness})
"""

    config = fspathtree(yaml.safe_load(config_text))

    config_renderer = rendering.ConfigRenderer()

    with pytest.raises(KeyError) as e:
        config = config_renderer.render(config)

    assert (
        e.value.args[0]
        == "Error while parsing path '/layer/0/thickness'. Could not find path element 'layer'."
    )


def test_creating_filenames():
    """
    Demonstrate how we can powerconf to generate filenames (for output for example) based on config parameters.
    """

    config_text = """
layers:
    - name: layer 1
      thickness: 10 um
      absorption: 2 1 / cm
grid:
    x:
      max: $( ${/layers/0/thickness})
laser:
    radius: 100 um
    irradiance: 2 W /cm^2

simulation: 
    output: Tvst-$(${/layers/0/absorption}.to("1/cm").magnitude|fmt(_1,".0f"))_inv_cm-$(${/laser/radius}.to("um").magnitude|fmt(_1,".0f"))_um.txt
    id: $( id(ctx))
"""

    config = fspathtree(yaml.safe_load(config_text))

    config_renderer = rendering.ConfigRenderer()

    def del_space(text):
        return text.replace(" ", "_")

    def to(q, u):
        return q.to(u).magnitude

    def fmt(v, spec=""):
        return ("{:" + spec + "}").format(v)

    assert fmt(10) == "10"
    config_renderer.expression_evaluator.add_global("del_space", del_space)
    config_renderer.expression_evaluator.add_global("fmt", fmt)
    config_renderer.expression_evaluator.add_global("to", to)
    config_renderer.expression_evaluator.add_global("id", utils.get_id)

    config = config_renderer.render(config)

    assert config["/simulation/output"] == "Tvst-2_inv_cm-100_um.txt"
    assert config["/simulation/id"] == "1c8c60bee71afc4f3143ef6edf3ca971"


def test_include_lists_basic(tmp_path):
    with unit_test_utils.working_directory(tmp_path):
        text1 = """
    simulation:
        grid:
            '@include' : 
                  - x-grid.yml
                  - y-grid.yml
    """

        text2 = """
    x: 
        min: 0 cm
        max: 1 cm
        N: 100
    """
        text3 = """
    y: 
        min: 1 cm
        max: 3 cm
        N: 200
        """

        pathlib.Path("x-grid.yml").write_text(text2)
        pathlib.Path("y-grid.yml").write_text(text3)

        config = loaders.yaml(text1)
        config = rendering.load_includes(config, loaders.yaml)
        # print(config.tree)

        assert "simulation/grid/x/min" in config
        assert "simulation/grid/x/max" in config
        assert "simulation/grid/x/N" in config
        assert "simulation/grid/y/min" in config
        assert "simulation/grid/y/max" in config
        assert "simulation/grid/y/N" in config

        assert config["simulation/grid/x/min"] == "0 cm"
        assert config["simulation/grid/x/max"] == "1 cm"
        assert config["simulation/grid/x/N"] == 100
        assert config["simulation/grid/y/min"] == "1 cm"
        assert config["simulation/grid/y/max"] == "3 cm"
        assert config["simulation/grid/y/N"] == 200


def test_include_lists_with_override(tmp_path):
    with unit_test_utils.working_directory(tmp_path):
        text1 = """
    simulation:
        grid:
            '@include' : 
                  - x-grid.yml
                  - y-grid.yml
    """

        text2 = """
    x: 
        min: 0 cm
        max: 1 cm
        N: 100
    """
        text3 = """
    x: 
        min: 0.5 cm
    y: 
        min: 1 cm
        max: 3 cm
        N: 200
        """

        pathlib.Path("x-grid.yml").write_text(text2)
        pathlib.Path("y-grid.yml").write_text(text3)

        config = loaders.yaml(text1)
        config = rendering.load_includes(config, loaders.yaml)
        # print(config.tree)

        assert "simulation/grid/x/min" in config
        assert "simulation/grid/x/max" in config
        assert "simulation/grid/x/N" in config
        assert "simulation/grid/y/min" in config
        assert "simulation/grid/y/max" in config
        assert "simulation/grid/y/N" in config

        assert config["simulation/grid/x/min"] == "0.5 cm"
        assert config["simulation/grid/x/max"] == "1 cm"
        assert config["simulation/grid/x/N"] == 100
        assert config["simulation/grid/y/min"] == "1 cm"
        assert config["simulation/grid/y/max"] == "3 cm"
        assert config["simulation/grid/y/N"] == 200


def test_include_at_root(tmp_path):
    with unit_test_utils.working_directory(tmp_path):
        text1 = """
    '@include': 'thermal.yml'
    simulation:
        grid:
          x:
            min: 0 cm
            max: 1 cm
            N: 100
    """

        text2 = """
    thermal: 
        rho: 1
        c: 2
        k: 3
    """

        pathlib.Path("thermal.yml").write_text(text2)

        config = loaders.yaml(text1)
        config = rendering.load_includes(config, loaders.yaml)

        assert "simulation/grid/x/min" in config
        assert "simulation/grid/x/max" in config
        assert "simulation/grid/x/N" in config
        assert "thermal/rho" in config
        assert "thermal/c" in config
        assert "thermal/k" in config

        assert config["simulation/grid/x/min"] == "0 cm"
        assert config["simulation/grid/x/max"] == "1 cm"
        assert config["simulation/grid/x/N"] == 100
        assert config["thermal/rho"] == 1
        assert config["thermal/c"] == 2
        assert config["thermal/k"] == 3


def test_include_at_root_with_list(tmp_path):
    with unit_test_utils.working_directory(tmp_path):
        text1 = """
    '@include':
      - 'thermal.yml'
      - 'optical.yml'
    simulation:
        grid:
          x:
            min: 0 cm
            max: 1 cm
            N: 100
    """

        text2 = """
    thermal: 
        rho: 1
        c: 2
        k: 3
    """
        text3 = """
    optical: 
        mu_a: 4
        mu_s: 5
    """

        pathlib.Path("thermal.yml").write_text(text2)
        pathlib.Path("optical.yml").write_text(text3)

        config = loaders.yaml(text1)
        config = rendering.load_includes(config, loaders.yaml)

        assert "simulation/grid/x/min" in config
        assert "simulation/grid/x/max" in config
        assert "simulation/grid/x/N" in config
        assert "thermal/rho" in config
        assert "thermal/c" in config
        assert "thermal/k" in config
        assert "optical/mu_a" in config
        assert "optical/mu_s" in config

        assert config["simulation/grid/x/min"] == "0 cm"
        assert config["simulation/grid/x/max"] == "1 cm"
        assert config["simulation/grid/x/N"] == 100
        assert config["thermal/rho"] == 1
        assert config["thermal/c"] == 2
        assert config["thermal/k"] == 3
        assert config["optical/mu_a"] == 4
        assert config["optical/mu_s"] == 5


def test_include_at_root_with_override(tmp_path):
    """
    parameters in the configuration file that is loaded should override
    any parameters in the included configurations.
    """
    with unit_test_utils.working_directory(tmp_path):
        text1 = """
    '@include':
      - 'thermal.yml'
      - 'optical.yml'
    simulation:
        grid:
          x:
            min: 0 cm
            max: 1 cm
            N: 100
    thermal:
        rho: 10
        v: 20
    """

        text2 = """
    thermal: 
        rho: 1
        c: 2
        k: 3
    """
        text3 = """
    optical: 
        mu_a: 4
        mu_s: 5
    """

        pathlib.Path("thermal.yml").write_text(text2)
        pathlib.Path("optical.yml").write_text(text3)

        config = loaders.yaml(text1)
        config = rendering.load_includes(config, loaders.yaml)

        assert "simulation/grid/x/min" in config
        assert "simulation/grid/x/max" in config
        assert "simulation/grid/x/N" in config
        assert "thermal/rho" in config
        assert "thermal/c" in config
        assert "thermal/k" in config
        assert "thermal/v" in config
        assert "optical/mu_a" in config
        assert "optical/mu_s" in config

        assert config["simulation/grid/x/min"] == "0 cm"
        assert config["simulation/grid/x/max"] == "1 cm"
        assert config["simulation/grid/x/N"] == 100
        assert config["thermal/rho"] == 10
        assert config["thermal/c"] == 2
        assert config["thermal/k"] == 3
        assert config["thermal/v"] == 20
        assert config["optical/mu_a"] == 4
        assert config["optical/mu_s"] == 5



import pint
import yaml
from fspathtree import fspathtree

from powerconf import utils

Q_ = pint.UnitRegistry().Quantity


def test_config_id():
    config1 = fspathtree({"one": 1})
    config2 = fspathtree({"one": 1, "two": 2})
    config3 = fspathtree({"one": 1})

    _id1 = utils.get_id(config1)
    _id2 = utils.get_id(config2)
    _id3 = utils.get_id(config3)

    assert _id1 != _id2
    assert _id1 == _id3
    assert _id2 != _id3

    assert utils.get_id(config2, path_predicate=lambda p: p.name != "two") == _id1


def test_config_id_with_quantities():
    config1 = fspathtree({"one": Q_(1, "cm")})
    config2 = fspathtree({"one": Q_(1, "cm"), "two": Q_(2, " cm")})
    config3 = fspathtree({"one": Q_(1, " cm")})

    _id1 = utils.get_id(config1)
    _id2 = utils.get_id(config2)
    _id3 = utils.get_id(config3)

    assert _id1 != _id2
    assert _id1 == _id3
    assert _id2 != _id3

    assert utils.get_id(config2, path_predicate=lambda p: p.name != "two") == _id1


def test_transforming_quantities_to_strings():
    config = fspathtree({"x": Q_(1, "cm"), "N": 1})
    config = utils.apply_transform(
        config, lambda p, n: str(n) if type(n) is Q_ else config[p]
    )

    assert config["x"] == "1 centimeter"
    assert config["N"] == 1


def test_transforming_quantities_to_strings_with_predicate():
    config = fspathtree({"x": Q_(1, "cm"), "N": 1})
    config = utils.apply_transform(
        config, lambda p, n: str(n), predicate=lambda p, n: hasattr(n, "magnitude")
    )

    assert config["x"] == "1 centimeter"
    assert config["N"] == 1


def test_apply_transform_to_list_of_configs():
    configs = []
    configs.append(fspathtree({"x": Q_(1, "cm"), "N": 1}))
    configs.append(fspathtree({"x": Q_(1, "cm"), "N": 1}))

    configs = utils.apply_transform(
        configs, lambda p, n: str(n), predicate=lambda p, n: hasattr(n, "magnitude")
    )

    assert len(configs) == 2
    assert configs[0]["x"] == "1 centimeter"
    assert configs[0]["N"] == 1
    assert configs[1]["x"] == "1 centimeter"
    assert configs[1]["N"] == 1


def test_retrieving_nodes_with_name():
    config = """
    grid:
      x:
        min: 0
        max: 1
        N: 100
      y:
        min: 1
        max: 3
        N: 200
    stretching:
      x:
       enabled: false
      y:
       enabled: true
    """
    config = fspathtree(yaml.safe_load(config))

    paths = list(map(str, config.get_all_paths(predicate=lambda p, n: p.name == "x")))

    assert len(paths) == 2
    assert "/grid/x" in paths
    assert "/stretching/x" in paths



import pyparsing
import pytest

from powerconf import parsing


def test_variable():
    result = parsing.variable.parse_string("$x")
    assert result["variable name"] == "x"
    result = parsing.variable.parse_string("$xyz")
    assert result["variable name"] == "xyz"
    result = parsing.variable.parse_string("$grid/x/min")
    assert result["variable name"] == "grid/x/min"
    result = parsing.variable.parse_string("$x_q")
    assert result["variable name"] == "x_q"

    with pytest.raises(pyparsing.exceptions.ParseException):
        parsing.variable.parse_string("x$x")

    with pytest.raises(pyparsing.exceptions.ParseException):
        parsing.variable.parse_string("$ x")

    result = parsing.variable.parse_string("${x}")
    assert result["variable name"] == "x"
    result = parsing.variable.parse_string("${/grid/x/min}")
    assert result["variable name"] == "/grid/x/min"
    result = parsing.variable.parse_string("${/grid/x/min val}")
    assert result["variable name"] == "/grid/x/min val"

    with pytest.raises(pyparsing.exceptions.ParseException):
        parsing.variable.parse_string("$ {x}")


def test_expressions():
    result = parsing.expression.parse_string("$(1 + 1)")
    assert result["expression body"] == "(1 + 1)"

    with pytest.raises(pyparsing.exceptions.ParseException):
        result = parsing.expression.parse_string("$ (1 + 1)")

    result = parsing.expression.parse_string("$( $x + $y)")
    assert result["expression body"] == "( $x + $y)"

    result = parsing.expression.parse_string("$( $x + ( 1 + 2 + $y))")
    assert result["expression body"] == "( $x + ( 1 + 2 + $y))"

    # result = parsing.expression_body.parse_string(result["expression body"])


def test_expressions_with_filters():
    result = parsing.expression.parse_string("$( $x + $y | to('cm') )")
    assert result["expression body"] == "( $x + $y | to('cm') )"

    code_and_filter = result["expression body"][1:-1].split("|")

    assert code_and_filter[0] == " $x + $y "
    assert code_and_filter[1] == " to('cm') "



import os

from powerconf import validation, yaml


def test_simple_validation(tmp_path):

    orig_path = os.getcwd()
    os.chdir(tmp_path)

    config_text = """
grid:
    x:
        min: 0
        max: 2
    y:
        max: 0
        min: 2
    """
    validation_text = """
def max_greater_than_min(config):
    assert config['/grid/x/max'] > config['/grid/x/min'], '/grid/x/max must be greater than /grid/x/min'
    assert config['/grid/y/max'] > config['/grid/y/min'], '/grid/y/max must be greater than /grid/y/min'
    """

    config_file = tmp_path / "CONFIG.yml"
    config_file.write_text(config_text)
    validation_file = tmp_path / "powerconf_validate.py"
    validation_file.write_text(validation_text)

    configs = yaml.powerload(config_file)

    results = validation.validate_config(configs[0], validation_file)

    assert len(results) == 1
    assert "max_greater_than_min" in results
    assert results["max_greater_than_min"]["result"] == "fail"
    assert "/grid/y/max must be greater than /grid/y/min" in str(
        results["max_greater_than_min"]["exception"]
    )

    os.chdir(orig_path)



import pint

# enable support for loading temperature quantities from a string.
# i.e. '37 degC'
ureg = pint.UnitRegistry(autoconvert_offset_to_baseunit=True)
Q_ = ureg.Quantity


def make_quantity(qstr):
    """
    Make a quantity from a string.

    For _most_ quantity strings we can just do Q_(qstr). However,
    it does not work for offset units.
    """
    v, u = qstr.split(maxsplit=1)
    v = float(v)
    u = u.strip()
    # handle inverse units that are given as the
    # value divided by a unit. i.e. "10 / s"
    # is "10 1/s"
    if len(u) > 0 and u[0] == "/":
        u = "1" + u
    return Q_(v, u)



class UndefinedVariableInExpression(RuntimeError):
    def __init__(self, msg):
        super().__init__(msg)


class SyntaxErrorInExpression(RuntimeError):
    def __init__(self, msg):
        super().__init__(msg)


class InvalidConfiguration(RuntimeError):
    def __init__(self, msg):
        super().__init__(msg)



import pytest

from powerconf import loaders, rendering


@pytest.mark.benchmark
def test_benchmark_chained_dependencies(benchmark):
    text = """
  a : 1
  b : $(${a}+1)
  c : $(${b}+1)
  d : $(${c}+1)
  e : $(${d}+1)
  f : $(${e}+1)
  g : $(${f}+1)
  h : $(${g}+1)
  i : $(${h}+1)
  j : $(${i}+1)
  k : $(${j}+1)
  """

    config = loaders.yaml(text)
    config_renderer = rendering.ConfigRenderer()
    config_renderer.render(config)
    assert config["b"] == "$(${a}+1)"

    benchmark(config_renderer.render, config)

    config = config_renderer.render(config)

    assert config["a"] == 1
    assert config["b"] == 2
    assert config["k"] == 11


@pytest.mark.benchmark
def test_benchmark_chained_dependencies_reverse(benchmark):
    text = """
  a : $(${b}+1)
  b : $(${c}+1)
  c : $(${d}+1)
  d : $(${e}+1)
  e : $(${f}+1)
  f : $(${g}+1)
  g : $(${h}+1)
  h : $(${i}+1)
  i : $(${j}+1)
  j : $(${k}+1)
  k : 1
  """

    config = loaders.yaml(text)
    config_renderer = rendering.ConfigRenderer()
    config_renderer.render(config)
    assert config["b"] == "$(${c}+1)"

    benchmark(config_renderer.render, config)

    config = config_renderer.render(config)

    assert config["a"] == 11
    assert config["b"] == 10
    assert config["k"] == 1



from typing import Annotated

from pydantic import (AfterValidator, AliasChoices, Field,
                      PlainSerializer, WithJsonSchema)

from .units import make_quantity

def QuantityWithUnit(U, names=None, desc=None):
    return Annotated[
        str,
        AfterValidator(lambda x: make_quantity(x).to(U)),
        PlainSerializer(lambda x: f"{x:~}" if x is not None else "null", return_type=str),
        WithJsonSchema({"type": "string"}, mode="serialization"),
        Field(
            alias=(
                AliasChoices(*names if isinstance(names, list) else [names])
                if names is not None
                else None
            ),
            description=desc,
        ),
    ]



import pint
import pytest
from fspathtree import fspathtree

from powerconf import expressions

ureg = pint.UnitRegistry()
Q_ = ureg.Quantity


def test_expression_evaluator():
    ctx = fspathtree()
    ctx["/grid/x/min"] = Q_(0, "cm")
    ctx["/grid/x/max"] = Q_(10, "cm")
    ctx["/grid/x/N"] = Q_(101, "")

    evaluator = expressions.ExecExpressionEvaluator()
    evaluator.start()
    evaluator.add_global("c", ctx)
    evaluator.add_global("Q_", Q_)
    dx = evaluator.eval("(c['/grid/x/max'] - c['/grid/x/min']) / (c['/grid/x/N']-1)")
    assert dx.magnitude == 0.1

    ctx["/grid/x/N"] = Q_(201, "")
    dx = evaluator.eval("(c['/grid/x/max'] - c['/grid/x/min']) / (c['/grid/x/N']-1)")
    assert dx.magnitude == 0.05

    evaluator.eval("numpy.exp( (c['/grid/x/max'] - c['/grid/x/min']) / Q_(10,'cm') )")
    assert dx.magnitude == 0.05

    evaluator.stop()


def test_expression_exceptions():
    with pytest.raises(expressions.ExpressionError) as e:
        raise expressions.ExpressionError("There was a problem.")

    assert e.value.args[0] == "There was a problem."

    with pytest.raises(expressions.ExpressionError) as e:
        try:
            exec("x = (a")
        except Exception as ee:
            raise expressions.ExpressionError(f"There was a problem: {ee}.")

    assert (
        e.value.args[0]
        == "There was a problem: '(' was never closed (<string>, line 1)."
    )


def test_expression_evaluator_bad_text():
    ctx = fspathtree()
    evaluator = expressions.ExecExpressionEvaluator()
    evaluator.start()
    evaluator.add_global("c", ctx)
    evaluator.add_global("Q_", Q_)

    with pytest.raises(RuntimeError) as e:
        evaluator.eval("import os")

    assert (
        e.value.args[0] == "Expressions are not allowed to contain the text 'import'."
    )


def test_expression_evaluator_with_filter():
    ctx = fspathtree()
    ctx["/grid/x/min"] = Q_(0, "cm")
    ctx["/grid/x/max"] = Q_(10, "cm")
    ctx["/grid/x/N"] = Q_(101, "")

    evaluator = expressions.ExecExpressionEvaluator()
    evaluator.start()
    evaluator.add_global("c", ctx)
    evaluator.add_global("Q_", Q_)
    evaluator.add_global("magnitude", lambda q: q.magnitude)
    evaluator.add_global("divide", lambda a, b: a / b)
    dx = evaluator.eval(
        "(c['/grid/x/max'] - c['/grid/x/min']) / (c['/grid/x/N']-1) | magnitude"
    )
    assert dx == 0.1
    dx = evaluator.eval(
        "(c['/grid/x/max'] - c['/grid/x/min']) / (c['/grid/x/N']-1) | magnitude(_1) | divide(2,_1)"
    )
    assert dx == 20

    evaluator.stop()


def test_expression_evaluator_with_exceptions():
    ctx = fspathtree()
    ctx["/grid/x/min"] = Q_(0, "cm")
    ctx["/grid/x/max"] = Q_(10, "cm")
    ctx["/grid/x/N"] = Q_(101, "")

    evaluator = expressions.ExecExpressionEvaluator()
    evaluator.start()
    evaluator.add_global("c", ctx)
    evaluator.add_global("Q_", Q_)
    evaluator.add_global("magnitude", lambda q: q.magnitude)
    evaluator.add_global("divide", lambda a, b: a / b)
    with pytest.raises(expressions.ExpressionError) as e:
        x = evaluator.eval("(1 + 2/3")
    assert e.value.args[0].startswith(
        "An exception was thrown while evaluating expression '(1 + 2/3'.\n"
    )
    assert "SyntaxError" in e.value.args[0]
    assert "'(' was never closed" in e.value.args[0]

    x = evaluator.eval("(1 + 2)/3")
    assert x == 1

    with pytest.raises(expressions.ExpressionError) as e:
        x = evaluator.eval("(1 + 2)/3 | missing")

    assert e.value.args[0].startswith(
        "An exception was thrown while evaluating expression ' missing(_configurator_xyz_result)' (generated by filter ' missing' in '(1 + 2)/3 | missing').\n"
    )
    assert "NameError" in e.value.args[0]
    assert "name 'missing' is not defined" in e.value.args[0]

    evaluator.stop()



import contextlib
import hashlib
import json
import os
import pathlib
import typing

from fspathtree import fspathtree


def get_id(
    config: fspathtree,
    path_predicate: typing.Callable[[fspathtree], bool] = None,
):
    """
    Return a unique id for the given configuration object.

    @param path_predicate: a function that will be passed the path to each node
    in the tree and should return true if the path sould be included in the
    tree when creating the id. This is useful for removing config parameters
    that are expected to change, but don't chnge the actual configuration.
    """
    if path_predicate is None:

        def path_predicate(p):
            return True

    # make a copy of the config with only keys not in the strip list
    c = fspathtree()
    for p in config.get_all_leaf_node_paths(predicate=path_predicate):
        c[p] = str(config[p])

    text = json.dumps(c.tree, sort_keys=True).replace(" ", "")
    return hashlib.md5(text.encode("utf-8")).hexdigest()


def apply_transform(
    config: fspathtree | typing.List[fspathtree],
    transform: typing.Callable[[fspathtree.PathType, typing.Any], typing.Any],
    predicate: typing.Callable[[fspathtree.PathType, typing.Any], typing.Any] = None,
):
    """
    Apply a transformation to configuration tree.

    Tree is modified in place.
    Transform is _ONLY_ applied to leaf nodes.


    @param transform: a function that accepts a path and node value, and returns the new value of the node.
    @param predicate: a function that accepst a path and node value, and returns True if the node should
                      have the transform apply. by default, all nodes have transform applied.
    """

    if type(config) is list:
        for i in range(len(config)):
            config[i] = apply_transform(config[i], transform, predicate)
    else:
        for p in config.get_all_leaf_node_paths(predicate=predicate):
            config[p] = transform(p, config[p])

    return config


@contextlib.contextmanager
def working_directory(path):
    path = pathlib.Path(path)
    path.mkdir(exist_ok=True, parents=True)
    last_dir = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(last_dir)
